@@ -5,6 +5,7 @@
  * Copyright (C) 2000-2003, Ximian, Inc.
  */
 
+#include <stdlib.h>
 #include <string.h>
 
 #include "soup-auth.h"
@@ -12,10 +13,10 @@
 #include "soup-message.h"
 #include "soup-message-private.h"
 #include "soup-misc.h"
+#include "soup-server-message.h"
 #include "soup-uri.h"
 
-#define PARENT_TYPE G_TYPE_OBJECT
-static GObjectClass *parent_class;
+G_DEFINE_TYPE (SoupMessage, soup_message, G_TYPE_OBJECT)
 
 enum {
 	WROTE_INFORMATIONAL,
@@ -45,12 +46,8 @@ static void finished (SoupMessage *req);
 static void free_chunks (SoupMessage *msg);
 
 static void
-init (GObject *object)
+soup_message_init (SoupMessage *msg)
 {
-	SoupMessage *msg = SOUP_MESSAGE (object);
-
-	msg->priv = g_new0 (SoupMessagePrivate, 1);
-
 	msg->status  = SOUP_MESSAGE_STATUS_IDLE;
 
 	msg->request_headers = g_hash_table_new (soup_str_case_hash,
@@ -59,18 +56,19 @@ init (GObject *object)
 	msg->response_headers = g_hash_table_new (soup_str_case_hash,
 						  soup_str_case_equal);
 
-	msg->priv->http_version = SOUP_HTTP_1_1;
+	SOUP_MESSAGE_GET_PRIVATE (msg)->http_version = SOUP_HTTP_1_1;
 }
 
 static void
 finalize (GObject *object)
 {
 	SoupMessage *msg = SOUP_MESSAGE (object);
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 
 	soup_message_io_stop (msg);
 
-	if (msg->priv->uri)
-		soup_uri_free (msg->priv->uri);
+	if (priv->uri)
+		soup_uri_free (priv->uri);
 
 	if (msg->request.owner == SOUP_BUFFER_SYSTEM_OWNED)
 		g_free (msg->request.body);
@@ -84,22 +82,20 @@ finalize (GObject *object)
 	soup_message_clear_headers (msg->response_headers);
 	g_hash_table_destroy (msg->response_headers);
 
-	g_slist_foreach (msg->priv->content_handlers, (GFunc) g_free, NULL);
-	g_slist_free (msg->priv->content_handlers);
+	g_slist_foreach (priv->content_handlers, (GFunc) g_free, NULL);
+	g_slist_free (priv->content_handlers);
 
 	g_free ((char *) msg->reason_phrase);
 
-	g_free (msg->priv);
-
-	G_OBJECT_CLASS (parent_class)->finalize (object);
+	G_OBJECT_CLASS (soup_message_parent_class)->finalize (object);
 }
 
 static void
-class_init (GObjectClass *object_class)
+soup_message_class_init (SoupMessageClass *message_class)
 {
-	SoupMessageClass *message_class = SOUP_MESSAGE_CLASS (object_class);
+	GObjectClass *object_class = G_OBJECT_CLASS (message_class);
 
-	parent_class = g_type_class_ref (PARENT_TYPE);
+	g_type_class_add_private (message_class, sizeof (SoupMessagePrivate));
 
 	/* virtual method definition */
 	message_class->wrote_body   = wrote_body;
@@ -113,6 +109,14 @@ class_init (GObjectClass *object_class)
 	object_class->finalize = finalize;
 
 	/* signals */
+
+	/**
+	 * SoupMessage::wrote-informational:
+	 * @msg: the message
+	 *
+	 * Emitted immediately after writing a 1xx (Informational)
+	 * response for a message.
+	 **/
 	signals[WROTE_INFORMATIONAL] =
 		g_signal_new ("wrote_informational",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -121,6 +125,13 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::wrote-headers:
+	 * @msg: the message
+	 *
+	 * Emitted immediately after writing the headers for a message.
+	 **/
 	signals[WROTE_HEADERS] =
 		g_signal_new ("wrote_headers",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -129,6 +140,14 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::wrote-chunk:
+	 * @msg: the message
+	 *
+	 * Emitted immediately after writing a body chunk for a message.
+	 * (This is
+	 **/
 	signals[WROTE_CHUNK] =
 		g_signal_new ("wrote_chunk",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -137,6 +156,13 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::wrote-body:
+	 * @msg: the message
+	 *
+	 * Emitted immediately after writing the complete body for a message.
+	 **/
 	signals[WROTE_BODY] =
 		g_signal_new ("wrote_body",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -146,6 +172,13 @@ class_init (GObjectClass *object_class)
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
+	/**
+	 * SoupMessage::got-informational:
+	 * @msg: the message
+	 *
+	 * Emitted after receiving a 1xx (Informational) response for
+	 * a message.
+	 **/
 	signals[GOT_INFORMATIONAL] =
 		g_signal_new ("got_informational",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -154,6 +187,13 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::got-headers:
+	 * @msg: the message
+	 *
+	 * Emitted after receiving all message headers for a message.
+	 **/
 	signals[GOT_HEADERS] =
 		g_signal_new ("got_headers",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -162,6 +202,16 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::got-chunk:
+	 * @msg: the message
+	 *
+	 * Emitted after receiving a chunk of a message body. Note
+	 * that "chunk" in this context means any subpiece of the
+	 * body, not necessarily the specific HTTP 1.1 chunks sent by
+	 * the other side.
+	 **/
 	signals[GOT_CHUNK] =
 		g_signal_new ("got_chunk",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -170,6 +220,13 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::got-body:
+	 * @msg: the message
+	 *
+	 * Emitted after receiving the complete message body.
+	 **/
 	signals[GOT_BODY] =
 		g_signal_new ("got_body",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -179,6 +236,12 @@ class_init (GObjectClass *object_class)
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
+	/**
+	 * SoupMessage::restarted:
+	 * @msg: the message
+	 *
+	 * Emitted when a message is about to be re-queued.
+	 **/
 	signals[RESTARTED] =
 		g_signal_new ("restarted",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -187,6 +250,15 @@ class_init (GObjectClass *object_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
+
+	/**
+	 * SoupMessage::finished:
+	 * @msg: the message
+	 *
+	 * Emitted when all HTTP processing is finished for a message.
+	 * (After #read-body for client-side code, or after
+	 * #wrote-body for server-side code.)
+	 **/
 	signals[FINISHED] =
 		g_signal_new ("finished",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -197,8 +269,6 @@ class_init (GObjectClass *object_class)
 			      G_TYPE_NONE, 0);
 }
 
-SOUP_MAKE_TYPE (soup_message, SoupMessage, class_init, init, PARENT_TYPE)
-
 
 /**
  * soup_message_new:
@@ -227,7 +297,7 @@ soup_message_new (const char *method, const char *uri_string)
 
 	msg = g_object_new (SOUP_TYPE_MESSAGE, NULL);
 	msg->method = method ? method : SOUP_METHOD_GET;
-	msg->priv->uri = uri;
+	SOUP_MESSAGE_GET_PRIVATE (msg)->uri = uri;
 
 	return msg;
 }
@@ -248,7 +318,7 @@ soup_message_new_from_uri (const char *method, const SoupUri *uri)
 
 	msg = g_object_new (SOUP_TYPE_MESSAGE, NULL);
 	msg->method = method ? method : SOUP_METHOD_GET;
-	msg->priv->uri = soup_uri_copy (uri);
+	SOUP_MESSAGE_GET_PRIVATE (msg)->uri = soup_uri_copy (uri);
 
 	return msg;
 }
@@ -695,7 +765,7 @@ soup_message_set_flags (SoupMessage *msg, guint flags)
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
-	msg->priv->msg_flags = flags;
+	SOUP_MESSAGE_GET_PRIVATE (msg)->msg_flags = flags;
 }
 
 /**
@@ -711,7 +781,7 @@ soup_message_get_flags (SoupMessage *msg)
 {
 	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), 0);
 
-	return msg->priv->msg_flags;
+	return SOUP_MESSAGE_GET_PRIVATE (msg)->msg_flags;
 }
 
 /**
@@ -728,7 +798,7 @@ soup_message_set_http_version (SoupMessage *msg, SoupHttpVersion version)
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
-	msg->priv->http_version = version;
+	SOUP_MESSAGE_GET_PRIVATE (msg)->http_version = version;
 }
 
 /**
@@ -745,7 +815,7 @@ soup_message_get_http_version (SoupMessage *msg)
 {
 	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), SOUP_HTTP_1_0);
 
-	return msg->priv->http_version;
+	return SOUP_MESSAGE_GET_PRIVATE (msg)->http_version;
 }
 
 /**
@@ -769,24 +839,27 @@ soup_message_is_keepalive (SoupMessage *msg)
 	    soup_method_get_id (msg->method) == SOUP_METHOD_ID_CONNECT)
 		return TRUE;
 
-	if (msg->priv->http_version == SOUP_HTTP_1_0) {
+	if (SOUP_MESSAGE_GET_PRIVATE (msg)->http_version == SOUP_HTTP_1_0) {
 		/* Only persistent if the client requested keepalive
 		 * and the server agreed.
 		 */
 
 		if (!c_conn || !s_conn)
 			return FALSE;
-		if (g_strcasecmp (c_conn, "Keep-Alive") != 0 ||
-		    g_strcasecmp (s_conn, "Keep-Alive") != 0)
+		if (g_ascii_strcasecmp (c_conn, "Keep-Alive") != 0 ||
+		    g_ascii_strcasecmp (s_conn, "Keep-Alive") != 0)
 			return FALSE;
 
 		return TRUE;
 	} else {
-		/* Persistent unless either side requested otherwise */
-
-		if (c_conn && g_strcasecmp (c_conn, "close") == 0)
+		/* Normally persistent unless either side requested otherwise */
+		if (c_conn && g_ascii_strcasecmp (c_conn, "close") == 0)
 			return FALSE;
-		if (s_conn && g_strcasecmp (s_conn, "close") == 0)
+		if (s_conn && g_ascii_strcasecmp (s_conn, "close") == 0)
+			return FALSE;
+
+		/* But not if the server sent a terminate-by-EOF response */
+		if (soup_message_get_response_encoding (msg, NULL) == SOUP_TRANSFER_EOF)
 			return FALSE;
 
 		return TRUE;
@@ -796,25 +869,28 @@ soup_message_is_keepalive (SoupMessage *msg)
 /**
  * soup_message_set_uri:
  * @msg: a #SoupMessage
- * @new_uri: the new #SoupUri
+ * @uri: the new #SoupUri
  *
  * Changes the URI that @msg is directed to (generally as a result
  * of a redirect).
  **/
 void
-soup_message_set_uri (SoupMessage *msg, const SoupUri *new_uri)
+soup_message_set_uri (SoupMessage *msg, const SoupUri *uri)
 {
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	SoupMessagePrivate *priv;
 
-	if (msg->priv->uri && new_uri) {
-		if (strcmp (msg->priv->uri->host, new_uri->host) != 0)
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	if (priv->uri && uri) {
+		if (strcmp (priv->uri->host, uri->host) != 0)
 			soup_message_io_stop (msg);
-	} else if (!new_uri)
+	} else if (!uri)
 		soup_message_io_stop (msg);
 
-	if (msg->priv->uri)
-		soup_uri_free (msg->priv->uri);
-	msg->priv->uri = soup_uri_copy (new_uri);
+	if (priv->uri)
+		soup_uri_free (priv->uri);
+	priv->uri = soup_uri_copy (uri);
 }
 
 /**
@@ -830,7 +906,130 @@ soup_message_get_uri (SoupMessage *msg)
 {
 	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), NULL);
 
-	return msg->priv->uri;
+	return SOUP_MESSAGE_GET_PRIVATE (msg)->uri;
+}
+
+/**
+ * soup_message_get_request_encoding:
+ * @msg: a #SoupMessage
+ * @content_length: a pointer to store the Content-Length in (or
+ * %NULL).
+ *
+ * Gets @msg's request encoding. For an outgoing (client) request,
+ * this is only valid after the message has been fully set up (from
+ * the library's perspective, that means not until the message has
+ * been queued). For an incoming (server) request, this is valid after
+ * the request headers have been read and @msg->request_headers filled
+ * in.
+ *
+ * Return value: the request encoding (which cannot be
+ * %SOUP_TRANSFER_UNKNOWN or %SOUP_TRANSFER_EOF). If it is
+ * %SOUP_TRANSFER_CONTENT_LENGTH, *@content_length will be set to the
+ * request body's length.
+ **/
+SoupTransferEncoding
+soup_message_get_request_encoding  (SoupMessage *msg, guint *content_length)
+{
+	if (SOUP_IS_SERVER_MESSAGE (msg)) {
+		const char *enc, *len;
+
+		enc = soup_message_get_header (msg->request_headers,
+					       "Transfer-Encoding");
+		len = soup_message_get_header (msg->request_headers,
+					       "Content-Length");
+		if (enc) {
+			if (g_ascii_strcasecmp (enc, "chunked") == 0)
+				return SOUP_TRANSFER_CHUNKED;
+			else
+				return SOUP_TRANSFER_UNKNOWN;
+		} else if (len) {
+			int lval = atoi (len);
+
+			if (lval < 0)
+				return SOUP_TRANSFER_UNKNOWN;
+			else {
+				if (content_length)
+					*content_length = lval;
+				return SOUP_TRANSFER_CONTENT_LENGTH;
+			}
+		} else
+			return SOUP_TRANSFER_NONE;
+	} else {
+		if (msg->request.length) {
+			if (content_length)
+				*content_length = msg->request.length;
+			return SOUP_TRANSFER_CONTENT_LENGTH;
+		} else
+			return SOUP_TRANSFER_NONE;
+	}
+}
+
+/**
+ * soup_message_get_response_encoding:
+ * @msg: a #SoupMessage
+ * @content_length: a pointer to store the Content-Length in (or
+ * %NULL).
+ *
+ * Gets @msg's response encoding. For an outgoing (client) request,
+ * this is only valid after the response headers have been read and
+ * @msg->response_headers filled in. For an incoming (server) request,
+ * this is valid after the server handler has run.
+ *
+ * Note that the returned value is the encoding actually used on the
+ * wire; this will not agree with the response headers in some cases
+ * (eg, a HEAD response may have a Content-Length header, but will
+ * still be considered %SOUP_TRANSFER_NONE by this function).
+ *
+ * Return value: the response encoding (which will not be
+ * %SOUP_TRANSFER_UNKNOWN). If it is %SOUP_TRANSFER_CONTENT_LENGTH,
+ * *@content_length will be set to the response body's length.
+ **/
+SoupTransferEncoding
+soup_message_get_response_encoding (SoupMessage *msg, guint *content_length)
+{
+	SoupMethodId method = soup_method_get_id (msg->method);
+
+	/* FIXME: should CONNECT really be here? Where does it say that? */
+	if (method == SOUP_METHOD_ID_HEAD ||
+	    method == SOUP_METHOD_ID_CONNECT ||
+	    msg->status_code  == SOUP_STATUS_NO_CONTENT ||
+	    msg->status_code  == SOUP_STATUS_NOT_MODIFIED ||
+	    SOUP_STATUS_IS_INFORMATIONAL (msg->status_code))
+		return SOUP_TRANSFER_NONE;
+
+	if (SOUP_IS_SERVER_MESSAGE (msg)) {
+		SoupTransferEncoding enc =
+			soup_server_message_get_encoding ((SoupServerMessage *)msg);
+		if (enc == SOUP_TRANSFER_UNKNOWN)
+			enc = SOUP_TRANSFER_CONTENT_LENGTH;
+		if (enc == SOUP_TRANSFER_CONTENT_LENGTH && content_length)
+			*content_length = msg->response.length;
+		return enc;
+	} else {
+		const char *enc, *len;
+
+		enc = soup_message_get_header (msg->response_headers,
+					       "Transfer-Encoding");
+		len = soup_message_get_header (msg->response_headers,
+					       "Content-Length");
+		if (enc) {
+			if (g_ascii_strcasecmp (enc, "chunked") == 0)
+				return SOUP_TRANSFER_CHUNKED;
+			else
+				return SOUP_TRANSFER_UNKNOWN;
+		} else if (len) {
+			int lval = atoi (len);
+
+			if (lval < 0)
+				return SOUP_TRANSFER_UNKNOWN;
+			else {
+				if (content_length)
+					*content_length = lval;
+				return SOUP_TRANSFER_CONTENT_LENGTH;
+			}
+		} else
+			return SOUP_TRANSFER_EOF;
+	}
 }
 
 /**
@@ -893,9 +1092,11 @@ soup_message_add_chunk (SoupMessage   *msg,
 			const char    *body,
 			guint          length)
 {
+	SoupMessagePrivate *priv;
 	SoupDataBuffer *chunk;
 
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	g_return_if_fail (body != NULL || length == 0);
 
 	chunk = g_new0 (SoupDataBuffer, 1);
@@ -908,11 +1109,11 @@ soup_message_add_chunk (SoupMessage   *msg,
 	}
 	chunk->length = length;
 
-	if (msg->priv->chunks) {
-		g_slist_append (msg->priv->last_chunk, chunk);
-		msg->priv->last_chunk = msg->priv->last_chunk->next;
+	if (priv->chunks) {
+		g_slist_append (priv->last_chunk, chunk);
+		priv->last_chunk = priv->last_chunk->next;
 	} else {
-		msg->priv->chunks = msg->priv->last_chunk =
+		priv->chunks = priv->last_chunk =
 			g_slist_append (NULL, chunk);
 	}
 }
@@ -944,17 +1145,19 @@ soup_message_add_final_chunk (SoupMessage *msg)
 SoupDataBuffer *
 soup_message_pop_chunk (SoupMessage *msg)
 {
+	SoupMessagePrivate *priv;
 	SoupDataBuffer *chunk;
 
 	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), NULL);
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 
-	if (!msg->priv->chunks)
+	if (!priv->chunks)
 		return NULL;
 
-	chunk = msg->priv->chunks->data;
-	msg->priv->chunks = g_slist_remove (msg->priv->chunks, chunk);
-	if (!msg->priv->chunks)
-		msg->priv->last_chunk = NULL;
+	chunk = priv->chunks->data;
+	priv->chunks = g_slist_remove (priv->chunks, chunk);
+	if (!priv->chunks)
+		priv->last_chunk = NULL;
 
 	return chunk;
 }
@@ -962,10 +1165,11 @@ soup_message_pop_chunk (SoupMessage *msg)
 static void
 free_chunks (SoupMessage *msg)
 {
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupDataBuffer *chunk;
 	GSList *ch;
 
-	for (ch = msg->priv->chunks; ch; ch = ch->next) {
+	for (ch = priv->chunks; ch; ch = ch->next) {
 		chunk = ch->data;
 
 		if (chunk->owner == SOUP_BUFFER_SYSTEM_OWNED)
@@ -973,6 +1177,6 @@ free_chunks (SoupMessage *msg)
 		g_free (chunk);
 	}
 
-	g_slist_free (msg->priv->chunks);
-	msg->priv->chunks = msg->priv->last_chunk = NULL;
+	g_slist_free (priv->chunks);
+	priv->chunks = priv->last_chunk = NULL;
 }
