@@ -19,7 +19,16 @@ struct {
 	{ "http://us%65r@host", "http://user@host" },
 	{ "http://us%40r@host", "http://us%40r@host" },
 	{ "http://us%3ar@host", "http://us%3ar@host" },
-	{ "http://us%2fr@host", "http://us%2fr@host" }
+	{ "http://us%2fr@host", "http://us%2fr@host" },
+
+	{ "http://control-chars/%01%02%03%04%05%06%07%08%09%0a%0b%0c%0d%0e%0f%10%11%12%13%14%15%16%17%18%19%1a%1b%1c%1d%1e%1f%7f",
+	  "http://control-chars/%01%02%03%04%05%06%07%08%09%0a%0b%0c%0d%0e%0f%10%11%12%13%14%15%16%17%18%19%1a%1b%1c%1d%1e%1f%7f"},
+	{ "http://space/%20",
+	  "http://space/%20" },
+	{ "http://delims/%3c%3e%23%25%22",
+	  "http://delims/%3c%3e%23%25%22" },
+	{ "http://unwise-chars/%7b%7d%7c%5c%5e%5b%5d%60",
+	  "http://unwise-chars/%7b%7d%7c%5c%5e%5b%5d%60" }
 };
 int num_abs_tests = G_N_ELEMENTS(abs_tests);
 
@@ -69,7 +78,14 @@ struct {
 	{ "g?y/../x", "http://a/b/c/g?y/../x" },
 	{ "g#s/./x", "http://a/b/c/g#s/./x" },
 	{ "g#s/../x", "http://a/b/c/g#s/../x" },
-	{ "http:g", "http:g" }
+
+	/* RFC 2396 notes that some old parsers will parse this as
+	 * a relative URL ("http://a/b/c/g"), but it should be
+	 * interpreted as absolute. libsoup should parse it
+	 * correctly as being absolute, but then reject it since it's
+	 * an http URL with no host.
+	 */
+	{ "http:g", NULL }
 };
 int num_rel_tests = G_N_ELEMENTS(rel_tests);
 
@@ -81,21 +97,33 @@ do_uri (SoupUri *base_uri, const char *base_str,
 	char *uri_string;
 
 	if (base_uri) {
-		printf ("<%s> + <%s> = <%s>? ", base_str, in_uri, out_uri);
+		printf ("<%s> + <%s> = <%s>? ", base_str, in_uri,
+			out_uri ? out_uri : "ERR");
 		uri = soup_uri_new_with_base (base_uri, in_uri);
 	} else {
-		printf ("<%s> => <%s>? ", in_uri, out_uri);
+		printf ("<%s> => <%s>? ", in_uri,
+			out_uri ? out_uri : "ERR");
 		uri = soup_uri_new (in_uri);
 	}
 
 	if (!uri) {
-		printf ("ERR\n  Could not parse %s\n", in_uri);
-		return FALSE;
+		if (out_uri) {
+			printf ("ERR\n  Could not parse %s\n", in_uri);
+			return FALSE;
+		} else {
+			printf ("OK\n");
+			return TRUE;
+		}
 	}
 
 	uri_string = soup_uri_to_string (uri, FALSE);
 	soup_uri_free (uri);
 
+	if (!out_uri) {
+		printf ("ERR\n  Got %s\n", uri_string);
+		return FALSE;
+	}
+
 	if (strcmp (uri_string, out_uri) != 0) {
 		printf ("NO\n  Unparses to <%s>\n", uri_string);
 		g_free (uri_string);
